from typing import NamedTuple, Optional
from pandas import DataFrame
from subprocess import run


from eazydocs.core._types import FunctionMethodType, StrPathType


class DfShape(NamedTuple):
    """Shape of a DataFrame.

    Attributes:
        rows (int): Number of rows.
        columns (int): Number of columns.
    """

    rows: int
    columns: int


class Example:
    def __init__(
        self,
        arg: DataFrame | FunctionMethodType,
        df_shape: Optional[tuple[int, int] | DfShape] = None,
        copy_to_clipboard: bool = True,
        append_to_method: Optional[str] = None,
        filename: Optional[StrPathType] = None,
        path: Optional[StrPathType] = None,
    ) -> None:
        """Generate an example representation of a DataFrame or method.

        Args:
            arg (DataFrame | FunctionMethodType): The DataFrame or method to
                generate an example for.
            df_shape (tuple[int,int] | DfShape, optional): A tuple or
                DfShape specifying the number of rows and columns to display
                from the DataFrame. If `type(arg)==DataFrame` and
                `df_shape=None`, the default shape of (5,5) will be used.
                Defaults to None.
            copy_to_clipboard (bool, optional): If True, the output will be
                copied to the clipboard. Defaults to True.
            append_to_method (str, optional): If provided, the example will be
                appended to the specified method in the markdown file. If
                `append_to_method!=None`, `filename` must also be provided.
                Optionally providing the `path` argument. Defaults to None.
            filename (StrPathType, optional): String or Path object for the
                filename. Defaults to None.
            path (StrPathType, optional): Directory path where the file will be
                located. Defaults to None.

        Raises:
            ValueError: If 'append_to_method' is provided but 'filename' is
                None.
        """
        if append_to_method and filename is None:
            raise ValueError(
                "'filename' must be provided if 'append_to_method' is used"
            )

        self.arg = arg
        self.copy_to_clipboard = copy_to_clipboard
        self.append_to_method = append_to_method
        self.filename = filename
        self.path = path

        if isinstance(arg, DataFrame):
            if df_shape is None:
                df_shape = (5, 5)
            output = self._format_df(df_shape)
        elif isinstance(arg, (FunctionMethodType)):
            output = self._format_method()
        else:
            raise TypeError("'arg' must be a DataFrame or method/function")

        if append_to_method is not None:
            raise NotImplementedError(
                "Appending to method not yet implemented"
            )

        if copy_to_clipboard:
            run(["clip.exe"], input=output.encode("utf-8"))
            print("Successfully copied example to clipboard!")

    def _format_df(self, shape: tuple[int, int]) -> str:
        """Provides a string representation of a DataFrame with specified shape.

        Args:
            shape (tuple[int,int]): A tuple specifying the number of rows and
                columns to display from the DataFrame.

        Raises:
            TypeError: If 'shape' is not a tuple or list of two integers.
            ValueError: If 'shape' does not contain exactly two integers.

        Returns:
            str: A formatted string representing the DataFrame.
        """
        if not isinstance(shape, tuple) and isinstance(shape, list):
            shape = tuple(shape)
        else:
            raise TypeError("'shape' must be a tuple or list of two integers")

        if len(shape) != 2:
            raise ValueError("'shape' must be a tuple of two integers")

        df = self.arg
        df = df.head(shape[0])
        df = df.iloc[:, 0 : shape[1]]

        output = df.to_string()
        output = "\n".join(f"\t{line}" for line in output.splitlines())

        return output

    def _format_method(self) -> str:
        """Provides a string representation of a method call with its arguments.

        Returns:
            str: A formatted string representing the method call.
        """
        output = f"```python\n{self.arg.__name__}()\n```"
        return output

    def _append_to_method(self, example: str) -> None:
        """Append the example to a specified method in a markdown file.

        Args:
            example (str): The example string to append.
        """
        from eazydocs.markdown.updater import Updater



from inspect import isclass
from pathlib import Path
from typing import Optional
from eazydocs.core._types import (
    ClassMethodType,
    FunctionMethodType,
    StrPathType,
)
from eazydocs.core.functions import create_md_file
from eazydocs.core.method import Method
from eazydocs.markdown.reader import Reader
from eazydocs.markdown.writer import Writer


class Updater:
    def __init__(
        self,
        class_or_method: ClassMethodType,
        filename: StrPathType,
        path: Optional[StrPathType] = None,
    ) -> None:
        """EazyDocs Updater class to update documentation for a class or method.

        Note:
            Providing a class will result in the file being overwritten since
            markdown files are expected to contain documentation for a
            single class and it's methods.

        Args:
            class_or_method (ClassMethodType): Class or method to update
                documentation for. If a class is provided, it will overwrite the
                provided file.
            filename (StrPathType): String or Path object representing the
                filename where the documentation is stored.
            path (StrPathType, optional): String or Path object for the path
                where the file will be saved. If not provided, the file will be
                saved in the current working directory. If provided, the
                `filename` will be joined to `path` argument. Defaults to None.
        """
        # Class provided, overwrite the file
        if isclass(class_or_method):
            return create_md_file(
                class_or_method=class_or_method,
                filename=filename,
                path=path,
                overwrite=True,
                update=True,
            )

        # Method provided, update the file
        filepath = self._valid_file(filename, path)
        self._update_md_file(filepath, class_or_method)

    def _update_md_file(
        self, filepath: Path, method: FunctionMethodType
    ) -> None:
        """Updates a method within the markdown file.

        Args:
            filepath (Path): Path object pointing to the markdown file to be
                updated.
            method (FunctionMethodType): Function or method to update
                documentation for.
        """
        method_docs = Method(method).parse(to_clipboard=False)

        contents, methods = self._read_md_file(filepath)

        old_method = methods.get(method.__name__)
        if old_method is None:
            # Append new method to the file
            contents = f"{contents}\n{method_docs}"
        else:
            self._update_method(
                contents, method.__name__, method_docs, filepath
            )

    def _valid_file(
        self,
        filename: StrPathType,
        path: Optional[StrPathType] = None,
    ) -> Path:
        """Check if the file is valid for updating.

        Args:
            filename (StrPathType): String or Path object representing the
                filename. Expected to be a markdown file (ends with .md).
            path (StrPathType, optional): The path where the file is located. If
                provided, it will be joined with the filename. Defaults to None.

        Returns:
            Path: A Path object pointing to the valid markdown file.

        Raises:
            ValueError: If the provided filename is invalid or does not exist at
                the specified path.
        """
        # Handle filename
        if isinstance(filename, str):
            filename = Path(filename)
        if not filename.suffix == ".md":
            filename = filename.with_suffix(".md")
        # Handle path
        if path is not None:
            filepath = Path(path) / filename
        else:
            filepath = Path(filename)

        if not filepath.exists():
            raise ValueError(
                f"Invalid file: '{filepath}'. Ensure the filename is spelled correctly and the path is correct."
            )
        return filepath

    def _read_md_file(self, filepath: Path) -> tuple[str, dict]:
        """Read the markdown file"""
        with Reader(filepath) as reader:
            contents = reader.contents
            methods = reader.methods
        return contents, methods

    def _update_method(
        self,
        current_docs: str,
        method_name: str,
        new_docs: str,
        filepath: Path,
    ) -> None:
        """Trim the old method documentation from the markdown file and
        replace it with the new documentation.

        Args:
            current_docs (str):  Current contents of the markdown file.
            method_name (str): Name of the method to be updated.
            new_docs (str): New documentation for the method.
            filepath (Path): Path object pointing to the markdown file to be
                updated.
        """
        # Format the method name to match the markdown format
        method_name = method_name.replace("_", "-")
        method_name = f"<strong id='{method_name}'"
        # Find the start and end of the old method
        start = current_docs.find(method_name)
        end = current_docs[start:].find("<hr>")
        # Get before and after old method docs
        before = current_docs[:start].strip()
        after = current_docs[(start + (end + 4)) :].strip()
        # Join parts
        updated_docs = f"{before}\n\n{new_docs}\n{after}"

        writer = Writer(updated_docs, filepath, overwrite=True, update=True)
        return writer.write()



from collections import defaultdict
import re
from subprocess import run
from types import FunctionType
from typing import Any, Optional
from eazydocs.core._parser import _Parser
from eazydocs.core._templates import (
    ARG_TEMPLATE,
    ARG_TEMPLATE_NODEFAULT,
    EXAMPLE_TEMPLATE,
)
from eazydocs.core._types import ArgsTuple, ParamsDict


class Method(metaclass=_Parser):
    """EazyDocs Method class.

    This class is used to parse the docstring of a method and extract
    information such as arguments, examples, and parameters. It is designed
    to be used with methods of a class, or functions, to generate documentation
    in a structured format.

    Upon initialization, it takes a method as an argument and parses its
    docstring to extract relevant information. The parsed information includes
    arguments, examples, and parameters, which can be accessed through the
    respective properties of the class.

    Attributes:
        method (object): The method to be parsed with docstring formatting
            defined by Google's style guide.

    Examples:
        >>> example_method.__doc__
        Example method for <class Example>.

        Args:
            param (list): Example required argument. If `param=None` then...
                else if `param=[]` then...
            param2 (str, optional): Example default None argument. Defaults to
                None.
            param3 (str, optional): Example default str argument. Defaults to
                "Test".

        Examples:

        General

            >>> d
            col1     col2
            0       First
            1      Second

        With DataFrame

            >>> data = Data("data.csv")
            >>> data.df
            Date_Time_1          vo2_1 vo2_2
            2025-01-01 00:00:00    0.0   0.0
            2025-01-01 01:00:00    0.0   0.0
            2025-01-01 02:00:00    0.0   0.0
            ...

        >>> Method(example_method).parse()
        <strong id='example'>Example</strong>(<b>param</b>=<i>None</i>, <b>param2</b>=<i>None</i>, <b>param3</b>=<i>"Test"</i>)

        Example method for <class Example>.

        > Parameters:

        <ul>
            <li>
                <b>param : <i>list, None</i></b>
                <ul style='list-style: none'>
                    <li>Example required argument. If <code>param=None</code> then... else if <code>param=[]</code> then...</li>
                </ul>
            </li>
            <li>
                <b>param2 : <i>str, optional, None</i></b>
                <ul style='list-style: none'>
                    <li>Example default None argument. Defaults to None.</li>
                </ul>
            </li>
            <li>
                <b>param3 : <i>str, optional, "Test"</i></b>
                <ul style='list-style: none'>
                    <li>Example default str argument. Defaults to "Test".</li>
                </ul>
            </li>
        </ul>

        > Examples:
        General

        ```python
        >>> d
        col1     col2
        0       First
        1      Second
        ```

        With DataFrame

        ```python
        >>> data = Data("data.csv")
        >>> data.df
        Date_Time_1          vo2_1 vo2_2
        2025-01-01 00:00:00    0.0   0.0
        2025-01-01 01:00:00    0.0   0.0
        2025-01-01 02:00:00    0.0   0.0
        ...
        ```
        <hr>
    """

    def __init__(self, method: object, include_examples: bool = True) -> None:
        """Initializes the Method instance.

        Args:
            method (object): The method to be parsed. It should have a docstring
                formatted according to Google's style guide.
            include_examples (bool, optional): Whether to include examples in
                the documentation. Defaults to True.
        """

        self.method = method
        self.docstring_split = self.docstring.split("\n")
        self.include_examples = include_examples

    def parse(self, to_clipboard: bool = True) -> str:
        """Parse the docstring of the method to extract arguments, examples,
        and parameters.

        Args:
            to_clipboard (bool, optional): If True, the output will be copied
                to the clipboard. Defaults to True.

        Returns:
            str: A formatted string containing the method's signature,
                summary, parameters, and examples (if `include_examples=True`).
        """
        if self.docstring is None:
            raise ValueError(
                f"{self.method} does not have a docstring, or is using an unsupported format. Ensure the method has a docstring formatted according to Google's style guide."
            )

        output = self._get_documentation()

        if to_clipboard:
            run(["clip.exe"], input=output.encode("utf-8"))
            print("Successfully copied to clipboard!")

        return output

    def _get_documentation(self) -> str:
        """Generates the documentation for the method."""
        output = f"{self.function}\n\n"
        output += f"{self.summary}\n\n"
        if self.args is not None:
            output += f"> Parameters:\n\n{self.args_fmtd}\n"
        if self.include_examples and self.examples is not None:
            output += f"> Examples:\n\n{self.examples}"

        output += "\n<hr>\n"

        return output

    def _get_args(self) -> list[ArgsTuple] | None:
        """Parse the arguments from the docstring.

        Returns:
            Optional[list[ArgsTuple]]: A list of tuples containing the
            arguments of the method, their types, and descriptions, or None if
            the docstring does not contain an 'Args' or 'Attributes' section.

        Notes:
            The returned args is a list of tuples, where each tuple contains:
                - param (str) - Name of the argument.
                - arg_type (str) - Type of the argument (e.g., int, str, list)
                - description (str) - Description of the argument.
        """
        if self.docstring is None:
            return None

        args_section = re.search(
            r"(Args|Attributes):\n(.*?)(\n\n|\Z)", self.docstring, re.DOTALL
        )
        if args_section is None:
            return None

        args_section = args_section.group(2)
        args_pattern = re.compile(
            r"\s*(\w+)\s*\(([\w\s,[\]\|]+)\):\s*(.*?)(?=\n\s*\w+\s*\(|\n\n|\Z)",
            re.DOTALL,
        )
        args = args_pattern.findall(args_section)
        return args

    def _get_examples(self) -> str | None:
        """Extracts examples from the docstring.

        Notes:
        - The Examples section is expected to be formatted as follows:
        ```
        Examples:
            Example 1

                >>> example_method(param1, param2)
                <optional output>

            Example 2

                >>> example_method(param1, param2)
                <optional output>
        ```
            - The formatted output will be wrapped in a template defined by the
                parameter's default value. See eazydocs.core._templates.py for
                templates.

        Returns:
            str | None: A string containing the formatted examples, or None if
                no examples are found in the docstring.
        """
        if self.docstring is None:
            return None

        pattern = re.compile(r"Example[s]?:.*?(?=\Z|Example:)", re.DOTALL)
        examples_section = pattern.search(self.docstring)
        if examples_section is None:
            return None

        examples_section = examples_section.group(0)

        pattern = re.compile(r"^\s{8}(\S.*)\n((\s{8}.*\n)*)", re.MULTILINE)
        matches = pattern.findall(examples_section)

        if matches is None:
            return None

        output = ""
        for match in matches:
            title = match[0].strip()
            example = "\n".join(
                line.strip() for line in match[1].strip().splitlines()
            )
            template = EXAMPLE_TEMPLATE.replace("{example}", example)
            output += f"{title}\n{template}\n\n"

        output = output.strip()

        return output

    def _get_function_signature(self) -> str | None:
        """Generates the function signature from the method's name and
        arguments."""
        if self.docstring is None:
            return None

        output = f"<strong id='{self.id}'>{self.name}</strong>("

        # Method has no params, close the signature and return
        if self.params is None:
            output += ")"
            return output

        # Otherwise, iterate through the params and add them to the signature
        for param in self.params:
            default_value = self.params[param].get("default_value")

            if default_value is None:
                output += f"<b>{param}</b>, "
            elif default_value == "None":
                output += f"<b>{param}</b>=<i>None</i>, "
            else:
                output += f"<b>{param}</b>=<i>{self.params[param].get("default_value")}</i>, "

        # Strip the trailing comma and space then add a closing parenthesis
        if output.endswith(", "):
            output = output[:-2] + ")"

        return output

    def _get_id(self) -> str:
        """Formats method name to create a unique ID."""
        id = self.name.replace("_", "-").lower()
        return id

    def _get_params(self) -> defaultdict[ParamsDict] | None:
        if self.args is None:
            return None

        params = defaultdict(ParamsDict)

        for arg in self.args:
            param, arg_type, description = arg
            description = self._format_description(description)
            default_value = self._get_default_value(description)

            params[param] = {
                "arg_type": arg_type,
                "description": description,
                "default_value": default_value,
            }

        return params

    def _get_summary(self) -> str | None:
        """Extracts the summary from the docstring.

        Returns:
            str|None: A string containing the summary of the method, or None
                if no summary is found in the docstring.
        """
        if self.docstring_split is None:
            return None

        summary = ""
        for line in self.docstring_split:
            if line.strip() in ["Args:", "Example:", "Equation:"]:
                break
            else:
                line = line.strip() + " "
                summary += line
        summary = summary.strip()

        return summary

    def _format_description(self, description: str) -> str:
        """Format the description of the argument by replace backticks and
            apostrophes with <code> tags.

        Args:
            description (str): String expression to format.
        """
        arg = re.sub(r"\s+", " ", description).strip()
        pattern = r"`"
        count = 0

        def replace(arg):
            nonlocal count
            count += 1
            return "<code>" if count % 2 != 0 else "</code>"

        return re.sub(pattern, replace, arg)

    def _get_default_value(self, arg: str) -> str:
        """Extracts the default value from the arg.

        Args:
            arg (str): Line from docstring to extract default value from.

        Returns:
            str: Default value of the argument or 'None' if
                no default value is found.
        """
        pattern = r"Defaults to (.+)\."
        match_obj = re.search(pattern, arg.strip())

        if match_obj:
            default_value = match_obj.group(1).strip()
            return f"{default_value}"
        else:
            default_value = "None"

        return default_value

    def _format_args(self) -> str:
        output = ""

        if self.params is None:
            return None

        for name, arg in self.params.items():
            default_arg = arg.get("default_value")

            if default_arg is None:
                template = ARG_TEMPLATE_NODEFAULT
                output += template.format(
                    method=self.name,
                    name=name,
                    arg_type=arg.get("arg_type"),
                    description=arg.get("description"),
                )
            else:
                template = ARG_TEMPLATE
                output += template.format(
                    method=self.name,
                    name=name,
                    arg_type=arg.get("arg_type"),
                    default_arg=default_arg,
                    description=arg.get("description"),
                )

        return f"<ul>\n{output}</ul>\n"

    # == Properties ==
    @property
    def args(self) -> list[Any]:
        return self._args

    @args.setter
    def args(self, val: list[Any]) -> None:
        self._args = val

    @property
    def args_fmtd(self) -> str:
        return self._args_fmtd

    @args_fmtd.setter
    def args_fmtd(self, val: str) -> None:
        self._args_fmtd = val

    @property
    def docstring(self) -> str:
        """The docstring of the method."""
        return self.method.__doc__ or ""

    @property
    def docstring_split(self) -> list[str] | None:
        """Split the docstring into a list of lines."""
        return self._docstring_split

    @docstring_split.setter
    def docstring_split(self, val: Optional[list[str]]) -> None:
        self._docstring_split = val

    @property
    def examples(self) -> str | None:
        """String containing examples from the docstring."""
        return self._examples

    @examples.setter
    def examples(self, val: str | None) -> None:
        self._examples = val

    @property
    def function(self) -> str:
        """Function signature for the method."""
        return self._function

    @function.setter
    def function(self, val: str) -> None:
        self._function = val

    @property
    def id(self) -> str:
        """The ID of the method, used for HTML element IDs."""
        return self.name.replace("_", "-").lower()

    @property
    def method(self) -> object:
        """The method object that this Method represents."""
        return self._method

    @method.setter
    def method(self, val: object) -> None:
        self._method = val

    @property
    def name(self) -> str:
        """The name of the method."""
        return self.method.__name__

    @name.setter
    def name(self, val: str) -> None:
        self._name = val

    @property
    def params(self) -> defaultdict[Any, ParamsDict]:
        return self._params

    @params.setter
    def params(self, val: list) -> None:
        self._params = val

    @property
    def summary(self) -> str | None:
        """Summary of the method extracted from the docstring."""
        return self._summary

    @summary.setter
    def summary(self, val: str | None) -> None:
        self._summary = val



from types import FunctionType, MethodType
from typing import Optional

from pandas import DataFrame
from eazydocs.core._types import (
    ClassMethodType,
    FunctionMethodType,
    StrPathType,
)
from eazydocs.core.example import DfShape, Example
from eazydocs.core.method import Method
from eazydocs.core.class_type import ClassType
from eazydocs.markdown.writer import Writer


def get_documentation(
    class_or_method: ClassMethodType,
    include_methods: bool = True,
    include_private_methods: bool = False,
    include_examples: bool = True,
    to_clipboard: bool = True,
) -> str:
    """Generate documentation for a class or method.

    Args:
        class_or_method (object | FunctionType | MethodType): The class or
            method to generate documentation for.
        include_methods (bool, optional): Whether to include methods in the
            documentation. Defaults to True.
        include_private_methods (bool, optional): Whether to include private
            methods in the documentation. Defaults to False.
        include_examples (bool, optional): Whether to include examples in the
            documentation. Defaults to True.
        to_clipboard (bool, optional): If True, the output will be copied to the
            clipboard. Defaults to True.
    """
    if isinstance(class_or_method, (FunctionType, MethodType)):
        method = Method(class_or_method, include_examples=include_examples)
        docs = method.parse(to_clipboard=to_clipboard)
    else:
        cls = ClassType(
            class_or_method,
            include_methods=include_methods,
            include_private_methods=include_private_methods,
            include_examples=include_examples,
        )
        docs = cls.parse(to_clipboard=to_clipboard)

    return docs


def create_md_file(
    class_or_method: ClassMethodType,
    filename: StrPathType = "README.md",
    path: Optional[StrPathType] = None,
    overwrite: bool = False,
    **kwargs,
) -> None:
    """Generate a markdown file for a class or method.

    Args:
        class_or_method (ClassMethodType): Class or method to generate
            documentation for.
        filename (StrPathType, optional): String or Path object for the
            filename. If a string object is provided, the file will be saved to
            the current working directory. Defaults to "README.md".
        path (StrPathType, optional): Directory path where the file will be
            saved. If not provided, the file will be saved in the current
            working directory. If provided, the `filename` will be joined to
            `path` argument. Defaults to None.
        overwrite (bool, optional): If True, the existing file will be
            overwritten without confirmation. Defaults to False.
    """
    docs = get_documentation(class_or_method, to_clipboard=False)

    writer = Writer(
        contents=docs,
        filename=filename,
        path=path,
        overwrite=overwrite,
        **kwargs,
    )

    writer.write()


def update_md_file(
    class_or_method: ClassMethodType,
    filename: StrPathType,
    path: Optional[StrPathType] = None,
) -> None:
    """Update an EazyDocs generated markdown file.

    If a class is provided, it will overwrite the provided file. Otherwise, it
    will trim the old method documentation from the file, insert the updated
    documentation, and write it to the given `filename`.

    Args:
        class_or_method (ClassMethodType): Class or method to update.
        filename (StrPathType): String or Path object representing the
            markdown file to update.
        path (StrPathType, optional): String or Path object for the path
            where the file will be saved. If not provided, the file will be
            saved in the current working directory. If provided, the
            `filename` will be joined to `path` argument. Defaults to None.
    """
    from eazydocs.markdown.updater import Updater

    Updater(class_or_method, filename, path)


def get_example(
    arg: DataFrame | FunctionMethodType,
    df_shape: Optional[tuple[int, int] | DfShape] = None,
    copy_to_clipboard: bool = True,
    append_to_method: Optional[str] = None,
    filename: Optional[StrPathType] = None,
    path: Optional[StrPathType] = None,
) -> None:
    """Generate an example representation of a DataFrame or method.

    Args:
        arg (DataFrame | FunctionMethodType): The DataFrame or method to
            generate an example for.
        df_shape (tuple[int,int] | DfShape, optional): A tuple or
            DfShape specifying the number of rows and columns to display
            from the DataFrame. If `type(arg)==DataFrame` and
            `df_shape=None`, the default shape of (5,5) will be used.
            Defaults to None.
        copy_to_clipboard (bool, optional): If True, the output will be
            copied to the clipboard. Defaults to True.
        append_to_method (str, optional): If provided, the example will be
            appended to the specified method in the markdown file. If
            `append_to_method!=None`, `filename` must also be provided.
            Optionally providing the `path` argument. Defaults to None.
        filename (StrPathType, optional): String or Path object for the
            filename. Defaults to None.
        path (StrPathType, optional): Directory path where the file will be
            located. Defaults to None.
    """
    Example(
        arg=arg,
        df_shape=df_shape,
        copy_to_clipboard=copy_to_clipboard,
        append_to_method=append_to_method,
        filename=filename,
        path=path,
    )



from pathlib import Path
from types import FunctionType, MethodType
from typing import NamedTuple, TypedDict, Union


class ParamsDict(TypedDict):
    arg_type: str
    description: str
    default_value: str


class ArgsTuple(NamedTuple):
    param: str
    arg_type: str
    description: str


FunctionMethodType = Union[FunctionType, MethodType]
ClassMethodType = Union[object, FunctionMethodType]
StrPathType = Union[str, Path]



"""eazydocs.core._templates.py

This module contains templates for generating documentation pages.
"""

ARG_TEMPLATE = """    <li>
        <b id='{method}-{name}'>{name} : <i>{arg_type}, {default_arg}</i></b>
        <ul style='list-style: none'>
            <li id='{method}-{name}-description'>{description}</li>
        </ul>
    </li>
"""

ARG_TEMPLATE_NODEFAULT = """    <li>
        <b id='{method}-{name}'>{name} : <i>{arg_type}</i></b>
        <ul style='list-style: none'>
            <li id='{method}-{name}-description'>{description}</li>
        </ul>
    </li>
"""


EXAMPLE_TEMPLATE = """
```python
{example}
```"""



from pathlib import Path
from typing import Literal, Optional
from eazydocs.core._types import StrPathType


class Writer:
    def __init__(
        self,
        contents: str,
        filename: StrPathType,
        path: Optional[StrPathType] = None,
        overwrite: bool = False,
        **kwargs,
    ) -> None:
        """Eazydocs Writer class to write markdown files.

        Args:
            contents (str): String object to write to the markdown file.
            filename (StrPathType): String or Path object for the filename.
            path (StrPathType, optional): String or Path object for the path
                where the file will be saved. If not provided, the file will be
                saved in the current working directory. If provided, the
                `filename` will be joined to `path` argument. Defaults to None.
            overwrite (bool, optional): Whether to overwrite the file if it
                already exists. Defaults to False.
        """
        self.contents = contents
        self.overwrite = overwrite

        self.filepath = self._get_filepath(filename, path)

        if kwargs:
            if kwargs.get("update") is not None:
                self.msg = (
                    f"Successfully updated markdown file: '{self.filepath}"
                )
        else:
            self.msg = f"Successfully created markdown file: '{self.filepath}'"

    def write(self) -> None:
        """Writes the `contents` to the markdown file at `filepath`, generated
        from the `filename` and `path`."""

        if not self._valid_path(self.filepath):
            return

        with open(self.filepath, "w+") as f:
            f.write(self.contents)

        print(self.msg)

    def _get_filepath(
        self,
        filename: StrPathType,
        path: Optional[StrPathType] = None,
    ) -> Path:
        """Get the filepath for the markdown file.

        Args:
            filename (StrPathType): String or Path object for the filename.
            path (StrPathType, optional): Directory path where the file will be
                saved. If not provided, the file will be saved in the current
                working directory. If provided, the `filename` will be joined to
                `path` argument. Defaults to None.

        Returns:
            Path: The Path object representing the filepath for the markdown file.
        """
        if isinstance(filename, str) and not filename.endswith(".md"):
            filename = f"{filename}.md"

        if path is not None:
            if isinstance(path, str):
                path = Path(path)
            filepath = Path(path) / filename
        else:
            if isinstance(filename, str):
                filepath = Path(filename)
            else:
                # Filename is already a Path object - Sanitize it
                if not filename.suffix:
                    filename = f"{filename}.md"
                filepath = filename

        return filepath

    def _valid_path(self, path: Path) -> bool:
        """Validate the filepath for the markdown file.

        If the filepath does not exist, prompt the user to create it. If the
        filepath exists and overwrite is False, prompt the user to confirm
        overwriting the file.

        Args:
            path: Path object representing the filepath to validate.

        Returns:
            bool: True if the filepath is valid, False otherwise.
        """
        if not path.exists():
            while True:
                confirm = input(
                    f"Filepath '{path}' does not exist. Do you want to create it? (y/n): "
                )
                match confirm.lower():
                    case "y":
                        print(f"Creating directory: '{path.parent}'")
                        break
                    case "n":
                        print("Exiting without creating the directory.")
                        return False
                    case _:
                        print(
                            "Invalid input. Please enter 'y' or 'n', or press Esc to exit."
                        )
        elif not self.overwrite:
            while True:
                confirm = input(
                    f"Filepath '{path}' already exists. Do you want to overwrite it? (y/n): "
                )
                match confirm.lower():
                    case "y":
                        print(f"Overwriting existing file: '{path}'")
                        break
                    case "n":
                        print("Exiting without overwriting the file.")
                        return False
                    case _:
                        print(
                            "Invalid input. Please enter 'y' or 'n', or press Esc to exit."
                        )

        path.parent.mkdir(parents=True, exist_ok=True)
        return True



from inspect import getmembers, isfunction, ismethod
from re import sub
from typing import Optional, Self
from eazydocs.core._parser import _Parser
from eazydocs.core.method import Method
from subprocess import run


class ClassType(metaclass=_Parser):
    """Eazydocs ClassType used to represent a class and its methods.

    This class is used to parse the docstring of a class and extract its
    methods, arguments, parameters, and examples. It is
    used in the Eazydocs documentation generation process.

    Args:
        cls (object): The class object to be parsed.
        include_methods (bool): Whether to include methods in the documentation.
            Defaults to True.
        include_private_methods (bool): Whether to include private methods in
            the documentation. Defaults to False.
        include_examples (bool): Whether to include examples in the
            documentation. Defaults to True.
        include_table_of_contents (bool): Whether to include a table of contents
            in the documentation. Defaults to True.

    Attributes:
        cls (object): The class object that this ClassType represents.
        docstring (str | None): The docstring of the class.
        id (str): The ID of the class, used for HTML element IDs.
        name (str): The name of the class.


    """

    def __init__(
        self,
        cls: object,
        include_methods: bool = True,
        include_private_methods: bool = False,
        include_examples: bool = True,
        include_table_of_contents: bool = True,
    ):
        self.include_methods = include_methods
        self.include_private_methods = include_private_methods
        self.include_examples = include_examples
        self.include_table_of_contents = include_table_of_contents

        self.init = None
        self.methods: list[Method] = []

    def parse(self, to_clipboard: bool = True) -> str:
        """This method parses the class and generates the documentation.
        Args:
            to_clipboard (bool): If True, the output will be copied to the
                clipboard. Defaults to True.

        Returns:
            str: The formatted documentation string for the class and its
                methods. If `to_clipboard` is True, the output will also be
                copied to the clipboard.
        """
        if self.output is None:
            self.__parse__()

        if to_clipboard:
            run(["clip.exe"], input=self.output.encode("utf-8"))
            print("Successfully copied to clipboard!")

        return self.output

    def __parse__(self) -> Self:
        for name, member in getmembers(self.cls):
            if ismethod(member) or isfunction(member):
                if self.include_methods:
                    if self.include_private_methods:
                        self.methods.append(
                            Method(member, self.include_examples)
                        )
                    elif not name.startswith("_"):
                        self.methods.append(
                            Method(member, self.include_examples)
                        )
                if name == "__init__":
                    self.init = member

        if self.init is not None:
            self.output = self._get_cls_documentation()

    def _get_cls_documentation(self) -> str:
        """Generates the documentation for the class.

        Returns:
            str: The formatted documentation string for the class and its
                methods.
        """
        method = self.init
        if method.__doc__ is None:
            return ""

        output = f"# {self.name}\n\n"

        if self.include_table_of_contents:
            output += self._get_table_of_contents()

        # Parse the docstring of the __init__ method
        docstring = Method(method)
        docstring.name = self.name

        # function_fmtd = docstring.function.replace(
        #     "--init--", self.name
        # ).replace("__init__", self.name)

        # Append the formatted function signature and summary
        output += f"{docstring.function}\n\n{docstring.summary}\n\n"
        # Append the examples if they exist
        if docstring.args:
            output += f"> Parameters:\n\n{docstring.args_fmtd}\n\n"

        output = sub(r"--init--|__init__", self.name, output)

        if docstring.examples:
            output += f"> Examples:\n\n{docstring.examples}\n\n"

        if self.methods != []:
            output += self._get_method_documentation()

        return output

    def _get_table_of_contents(self) -> str:
        """Generates the table of contents providing links to methods."""
        table_of_contents = f"- [Class Documentation](#{self.id})\n"

        for method in self.methods:
            table_of_contents += f"- [{method.name}](#{method.id})\n"
        table_of_contents += "\n"
        return table_of_contents

    def _get_method_documentation(self) -> str:
        """Generates the documentation for the class methods."""
        output = "## Class Methods\n\n"
        for method in self.methods:
            output += method._get_documentation()
        return output

    @property
    def cls(self) -> object:
        """The class object that this ClassType represents."""
        return self._cls

    @cls.setter
    def cls(self, val: object) -> None:
        self._cls = val

    @property
    def docstring(self) -> str | None:
        """The docstring of the class."""
        return self.cls.__init__.__doc__ or self.cls.__doc__ or None

    @property
    def id(self) -> str:
        """The ID of the class, used for HTML element IDs."""
        return self.name.replace("_", "-").lower()

    @property
    def init(self) -> Method | None:
        """The __init__ method of the class, if it exists."""
        return self._init

    @init.setter
    def init(self, val: Optional[Method] = None) -> None:
        self._init = val

    @property
    def methods(self) -> list[Method]:
        """The list of methods in the class."""
        return self._methods

    @methods.setter
    def methods(self, val: list[Method]) -> None:
        self._methods = val

    @property
    def name(self) -> str:
        """The name of the class."""
        return self.cls.__name__

    @property
    def output(self) -> str:
        """The formatted output of the class documentation."""
        return self._output

    @output.setter
    def output(self, val: str) -> None:
        self._output = val



from pathlib import Path
from re import findall
from typing import Literal, Optional


class Reader:
    """EazyDocs Markdown Reader class to read markdown files and parse their
    contents.

    Notes:
        - The markdown file is expected to have been generated by EazyDocs.
    """

    def __init__(self, path: Path):
        """EazyDocs Markdown Reader class to read markdown files and parse their contents.

        Args:
            path (Path): Path object pointing to the markdown file to be read.
        """
        self.path = path

        self.methods = dict()

    def __enter__(self, path: Optional[Path] = None):
        """Context manager interface for the Reader class. Reads the markdown
        file and prepares it for parsing.

        1. Reads the markdown file from the provided path.
        2. Extracts methods from the markdown file contents.



        Args:
            path (Optional[Path], optional): Path to the markdown file to be
                read. If not provided, uses the path set during initialization.
                Defaults to None.
        """
        if path is None:
            path = self.path

        with open(path, "r", encoding="utf-8") as file:
            self.contents = file.read()

        self._get_methods()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> Literal[False]:
        """Context manager interface for the Reader class."""
        return False

    def _get_methods(self) -> list[str] | None:
        pattern = r"\<strong .+\>(.+)</strong>"
        methods = findall(pattern, self.contents)
        if methods == []:
            return None

        for method in methods:
            params = self._parse_params(method)
            if params is not None:
                self.methods.update({method: params})

    def _parse_params(self, method: str) -> dict | None:
        """Parses the parameters of a method from the markdown contents.

        Notes:
            - This method extracts parameter names and descriptions for a given
              method from the markdown contents.
            - It uses regular expressions to find parameter names and their
              corresponding descriptions.
                - Markdown files generated by EazyDocs follow a specific format
                  where parameters are defined as `method-paramName'` and their
                  descriptions as `method-paramName-description'.`

        Args:
            method (str): The method name to parse parameters for.

        Returns:
            dict | None: A dictionary containing parameter names as keys and
                their descriptions as values. Returns None if no parameters are
                found for the method.
        """
        # Get parameter names
        pattern = rf"{method}\-(.*)\'"
        param_names = findall(pattern, self.contents)
        param_names = [p for p in param_names if "-description" not in p]
        # Return None if no parameters found
        if param_names == []:
            return None
        # Get parameter descriptions and build dictionary
        params = dict()

        for param in param_names:
            pattern = rf"{param}-description'>(.+)\."
            description = findall(pattern, self.contents)

            if description != []:
                params.update({param: description[0]})
            else:
                params.update({param: None})
        return params

    @property
    def contents(self) -> str:
        """Contents of the markdown file."""
        return self._contents

    @contents.setter
    def contents(self, val: str) -> None:
        self._contents = val

    @property
    def methods(self) -> dict[str, dict[str, str]]:
        """Methods parsed from the markdown file.

        Returns:
            dict: A dictionary where keys are method names
                and values are dictionaries of parameter names and their
                descriptions.
        """
        return self._methods

    @methods.setter
    def methods(self, val: dict[str, dict[str, str]]) -> None:
        self._methods = val



class _Parser(type):
    """Eazydocs Parser Metaclass utilized to parse methods and their docstrings.

    This metaclass is responsible for parsing the docstring of a method to
    extract arguments, parameters, examples, and formatted arguments. It is
    used in the MethodType class to create an instance with the parsed method.
    """

    def __call__(self, *args, **kwargs):
        """The metaclass for the ClassType and MethodType classes.

        This metaclass is used to create a new instance of the class and set the
        attributes based on the provided keyword arguments.

        A single argument is expected, which is the class/method to be parsed.

        Raises:
            TypeError: If no arguments are provided to the metaclass.

        Returns:
            ClassType|MethodType: An instance of the class or method type
            with parsed attributes.
        """
        if args is None:
            raise TypeError("Parser requires at least one argument.")

        # Get the first argument, which should be a class or method
        arg = args[0]

        from eazydocs.core.method import Method
        from eazydocs.core.class_type import ClassType

        instance = super().__call__(*args, **kwargs)

        if isinstance(instance, Method):
            instance.summary = instance._get_summary()
            instance.args = instance._get_args()
            instance.params = instance._get_params()
            instance.function = instance._get_function_signature()
            instance.examples = instance._get_examples()
            instance.args_fmtd = instance._format_args()
        elif isinstance(instance, ClassType):
            instance.cls = arg
            instance.__parse__()

        return instance
